-- ==============================================================
-- FILE : HSB_MSTR.VHD  -  can be modified by the USER
-- ==============================================================
--
-- This file is part of the HSB Mastering example for FPGA modules.
--
-- This file can be modified by users to implement their own logic.
--
-- ==============================================================
--
--  Module      : HSB_MSTR
--  Date        : 24/10/2003
--  Author      : R. Williams - HUNT ENGINEERING
--  Description : Using HUNT Message Interface to send messages
--
-- ==============================================================
--
--  Ver     Modified By      Date      Changes
--  ---     -----------      ----      -------
--  1.0     R. Williams    24-10-03    First Written
--
-- ==============================================================


LIBRARY ieee;
  USE ieee.std_logic_1164.ALL;
  USE ieee.std_logic_arith.ALL;
  USE ieee.std_logic_unsigned.ALL;

--pragma translate_off
LIBRARY unisim;
  USE unisim.vcomponents.ALL;
--pragma translate_on


entity HSB_MSTR is
  port (
    RST          : in  std_logic;
    CLK          : in  std_logic;
    SEND_ONE_MSG : in  std_logic;
    RESTART      : in  std_logic;
    MSG_SENT     : out std_logic;
    END_OF_MSGS  : out std_logic;
    DONE         : in  std_logic;
    COUNT        : in  std_logic;
    CLEAR        : in  std_logic;
    SEND_MSG     : out std_logic;
    CE           : out std_logic;
    DOUT         : out std_logic_vector(7 downto 0);
    SEND_ID      : out std_logic;
    LAST_BYTE    : out std_logic
  );
end HSB_MSTR;


architecture RTL of HSB_MSTR is

  component MSG_RAM
    port (
      RST  : in  std_logic;
      CLK  : in  std_logic;
      ADDR : in  std_logic_vector(9 downto 0);
      DI   : in  std_logic_vector(15 downto 0);
      DO   : out std_logic_vector(15 downto 0)
    );
  end component;

  signal RESET           : std_logic;
  signal SEND_MSG_int    : std_logic;
  signal END_OF_MSGS_int : std_logic;
  signal MSG_SENT_int    : std_logic;

  signal RAM_ADDR        : std_logic_vector(9 downto 0);
  signal RAM_DI          : std_logic_vector(15 downto 0);
  signal RAM_DO          : std_logic_vector(15 downto 0);

  signal DELAY_ce        : std_logic;
  signal DELAY_count     : std_logic_vector(10 downto 0);
  signal DELAY_tc        : std_logic_vector(10 downto 0);

  signal CE_int          : std_logic;
  signal CE_count        : std_logic_vector(1 downto 0);

begin

  SEND_MSG    <= SEND_MSG_int;
  END_OF_MSGS <= END_OF_MSGS_int;
  MSG_SENT    <= MSG_SENT_int;

  -- Reset all logic when the system reset is asserted or the RESTART signal is asserted
  RESET <= RST OR RESTART;

  -- Cycle through the process of sending a message each time SEND_ONE_MSG is set to 1
  process (RESET, CLK)
  begin
    if RESET='1' then
      SEND_MSG_int <= '0';
    elsif rising_edge(CLK) then
      SEND_MSG_int <= '0';
      if (SEND_MSG_int='0' and SEND_ONE_MSG='1' and END_OF_MSGS_int='0') or
         (SEND_MSG_int='1' and CLEAR='0') then
        SEND_MSG_int <= '1';
      end if;
    end if;
  end process;

  -- Address generation for the Block RAM. Increment address when COUNT is asserted
  process(RESET, CLK)
  begin
    if RESET='1' then
      RAM_ADDR <= (others=>'0');
    elsif rising_edge(CLK) then
      if COUNT='1' then
        RAM_ADDR <= RAM_ADDR + 1;
      end if;
    end if;
  end process;

  -- Add a delay between the end of one message and transmission of the next
  process(RESET, CLK)
  begin
    if RESET='1' then
      DELAY_ce <= '0';
    elsif rising_edge(CLK) then
      if CE_int='1' then
        DELAY_ce <= '0';
        if (DELAY_ce='0' and DONE='1') or
           (DELAY_ce='1' and MSG_SENT_int='0') then
          DELAY_ce <= '1';
        end if;
      end if;
    end if;
  end process;

  DELAY_tc <= "11111111111";

  MSG_SENT_int <= '1' when DELAY_count=DELAY_tc else '0';

  process(RESET, CLK)
  begin
    if RESET='1' then
      DELAY_count <= (others=>'0');
    elsif rising_edge(CLK) then
      if DELAY_ce='1' then
        DELAY_count <= DELAY_count + 1;
      else
        DELAY_count <= (others=>'0');
      end if;
    end if;
  end process;

  -- Logic to control the CE signal of the HE_USER Message Interface
  -- Assert CE every four clock cycles
  CE_int <= '1' when CE_count="11" else '0';

  CE <= CE_int;

  process(CLK)
  begin
    if rising_edge(CLK) then
      CE_count <= CE_count + 1;
    end if;
  end process;

  -- Block RAM containing all of the data for all messages to be sent
  iMSG_RAM : MSG_RAM
    port map (
      RST  => RST,
      CLK  => CLK,
      ADDR => RAM_ADDR,
      DI   => RAM_DI,
      DO   => RAM_DO );

  RAM_DI <= (others=>'0');

  -- Data byte output over HSB
  DOUT      <= RAM_DO(7 downto 0);

  -- Control signal to indicate the module's own ID must be transmitted
  SEND_ID   <= RAM_DO(8);

  -- Control signal to indicate the current byte is the last byte of the message
  LAST_BYTE <= RAM_DO(9);

  -- Control signal to indicate the end of messages in the Block RAM
  END_OF_MSGS_int <= RAM_DO(10);

end RTL;
