-- ==============================================================
-- FILE : GD11.VHD
-- ==============================================================
--
-- This file provides example VHDL code for using your FPGA
-- module with a HEGD11.
--
-- Copy and Paste the lines of VHDL from this file into your
-- own project.
--
-- ==============================================================
--
--  Date        : 15/09/2004
--  Author      : R. Williams - HUNT ENGINEERING
--  Description : GD11 Example Code
--
-- ==============================================================
--
--  Ver     Modified By      Date      Changes
--  ---     -----------      ----      -------
--  1.0     R. Williams    15-09-04    First Written
--
-- ==============================================================


  -- The following code must appear in the VHDL architecture header:

  -- Data Enable Signal Declarations
  signal GD11_CH0_EN  : std_logic;
  signal GD11_CH1_EN  : std_logic;
  signal GD11_CH2_EN  : std_logic;
  signal GD11_CH3_EN  : std_logic;
  signal GD11_CH4_EN  : std_logic;
  signal GD11_CH5_EN  : std_logic;
  signal GD11_CH6_EN  : std_logic;
  signal GD11_CH7_EN  : std_logic;
  signal GD11_CH8_EN  : std_logic;
  signal GD11_CH9_EN  : std_logic;
  signal GD11_CH10_EN : std_logic;
  signal GD11_CH11_EN : std_logic;

  -- Internal Read Request signal
  signal READ_REQ : std_logic;

  -- Unsigned Data Bus Signal Declaration
  signal GD11_UNS_DATA : std_logic_vector(11 downto 0);

  -- Signed Data Bus Signal Declaration
  signal GD11_SGN_DATA : std_logic_vector(11 downto 0);

  -- INFIFO_DVALID(0) delayed by one clock cycle
  signal DVALID_DELAYED : std_logic;

  -- Bus to store top half of input data word
  signal TOP_HALF : std_logic_vector(15 downto 0);


  -- The following code must appear in the VHDL architecture body.

  -- PLEASE NOTE. All signals defined below must be used in the HERON
  -- Input FIFO Clock Domain. That is, all are clocked using the signal
  -- FCLK_G.

  -- For this example VHDL, HERON Input FIFO 0 is connected to the data
  -- stream sourced by the HEGD11.

  -- The signals INFIFO_READ_REQ, INFIFO_DVALID and INFIFO0_D are used.
  -- These signals are found in the port list of the USER-AP entity in
  -- your project.

  -- Toggle the Input FIFO Read Request to limit one word arriving
  -- no more than every two clock cycles. Doing this creates enough
  -- time to unpack the top and bottom of the input data word.
  process(RESET, FCLK_G)
  begin
    if RESET='1' then
      READ_REQ <= '0';
    elsif rising_edge(FCLK_G) then
      READ_REQ <= '0';
      if READ_REQ='0' then
        READ_REQ <= '1';
      end if;
    end if;
  end process;

  INFIFO_READ_REQ <= "00000" & READ_REQ;

  -- Delay the Input FIFO data valid signal by one clock cycle
  process(RESET, FCLK_G)
  begin
    if RESET='1' then
      DVALID_DELAYED <= '0';
    elsif rising_edge(FCLK_G) then
      DVALID_DELAYED <= INFIFO_DVALID(0);
    end if;
  end process;

  -- Register the top half of the Input FIFO data word
  process(RESET, FCLK_G)
  begin
    if RESET='1' then
      TOP_HALF <= (others=>'0');
    elsif rising_edge(FCLK_G) then
      if INFIFO_DVALID(0)='1' then
        TOP_HALF <= INFIFO0_D(31 downto 16);
      end if;
    end if;
  end process;

  -- The HEGD11 provides 12-bit unsigned data

  -- To take data from HEGD11 Channel 0, drive the data enable of your
  -- first stage of processing with the signal GD11_CH0_EN.
  GD11_CH0_EN <= (INFIFO_DVALID(0) AND (NOT INFIFO0_D(15)) AND (NOT INFIFO0_D(14)) AND (NOT INFIFO0_D(13)) AND (NOT INFIFO0_D(12))) OR
                 (DVALID_DELAYED   AND (NOT  TOP_HALF(15)) AND (NOT  TOP_HALF(14)) AND (NOT  TOP_HALF(13)) AND (NOT  TOP_HALF(12)));

  -- To take data from HEGD11 Channel 1, drive the data enable of your
  -- first stage of processing with the signal GD11_CH1_EN.
  GD11_CH1_EN <= (INFIFO_DVALID(0) AND (NOT INFIFO0_D(15)) AND (NOT INFIFO0_D(14)) AND (NOT INFIFO0_D(13)) AND (    INFIFO0_D(12))) OR
                 (DVALID_DELAYED   AND (NOT  TOP_HALF(15)) AND (NOT  TOP_HALF(14)) AND (NOT  TOP_HALF(13)) AND (     TOP_HALF(12)));

  -- To take data from HEGD11 Channel 2, drive the data enable of your
  -- first stage of processing with the signal GD11_CH2_EN.
  GD11_CH2_EN <= (INFIFO_DVALID(0) AND (NOT INFIFO0_D(15)) AND (NOT INFIFO0_D(14)) AND (    INFIFO0_D(13)) AND (NOT INFIFO0_D(12))) OR
                 (DVALID_DELAYED   AND (NOT  TOP_HALF(15)) AND (NOT  TOP_HALF(14)) AND (     TOP_HALF(13)) AND (NOT  TOP_HALF(12)));

  -- To take data from HEGD11 Channel 3, drive the data enable of your
  -- first stage of processing with the signal GD11_CH3_EN.
  GD11_CH3_EN <= (INFIFO_DVALID(0) AND (NOT INFIFO0_D(15)) AND (NOT INFIFO0_D(14)) AND (    INFIFO0_D(13)) AND (    INFIFO0_D(12))) OR
                 (DVALID_DELAYED   AND (NOT  TOP_HALF(15)) AND (NOT  TOP_HALF(14)) AND (     TOP_HALF(13)) AND (     TOP_HALF(12)));

  -- To take data from HEGD11 Channel 4, drive the data enable of your
  -- first stage of processing with the signal GD11_CH4_EN.
  GD11_CH4_EN <= (INFIFO_DVALID(0) AND (NOT INFIFO0_D(15)) AND (    INFIFO0_D(14)) AND (NOT INFIFO0_D(13)) AND (NOT INFIFO0_D(12))) OR
                 (DVALID_DELAYED   AND (NOT  TOP_HALF(15)) AND (     TOP_HALF(14)) AND (NOT  TOP_HALF(13)) AND (NOT  TOP_HALF(12)));

  -- To take data from HEGD11 Channel 5, drive the data enable of your
  -- first stage of processing with the signal GD11_CH5_EN.
  GD11_CH5_EN <= (INFIFO_DVALID(0) AND (NOT INFIFO0_D(15)) AND (    INFIFO0_D(14)) AND (NOT INFIFO0_D(13)) AND (    INFIFO0_D(12))) OR
                 (DVALID_DELAYED   AND (NOT  TOP_HALF(15)) AND (     TOP_HALF(14)) AND (NOT  TOP_HALF(13)) AND (     TOP_HALF(12)));

  -- To take data from HEGD11 Channel 6, drive the data enable of your
  -- first stage of processing with the signal GD11_CH6_EN.
  GD11_CH6_EN <= (INFIFO_DVALID(0) AND (NOT INFIFO0_D(15)) AND (    INFIFO0_D(14)) AND (    INFIFO0_D(13)) AND (NOT INFIFO0_D(12))) OR
                 (DVALID_DELAYED   AND (NOT  TOP_HALF(15)) AND (     TOP_HALF(14)) AND (     TOP_HALF(13)) AND (NOT  TOP_HALF(12)));

  -- To take data from HEGD11 Channel 7, drive the data enable of your
  -- first stage of processing with the signal GD11_CH7_EN.
  GD11_CH7_EN <= (INFIFO_DVALID(0) AND (NOT INFIFO0_D(15)) AND (    INFIFO0_D(14)) AND (    INFIFO0_D(13)) AND (    INFIFO0_D(12))) OR
                 (DVALID_DELAYED   AND (NOT  TOP_HALF(15)) AND (     TOP_HALF(14)) AND (     TOP_HALF(13)) AND (     TOP_HALF(12)));

  -- To take data from HEGD11 Channel 8, drive the data enable of your
  -- first stage of processing with the signal GD11_CH8_EN.
  GD11_CH8_EN <= (INFIFO_DVALID(0) AND (    INFIFO0_D(15)) AND (NOT INFIFO0_D(14)) AND (NOT INFIFO0_D(13)) AND (NOT INFIFO0_D(12))) OR
                 (DVALID_DELAYED   AND (     TOP_HALF(15)) AND (NOT  TOP_HALF(14)) AND (NOT  TOP_HALF(13)) AND (NOT  TOP_HALF(12)));

  -- To take data from HEGD11 Channel 9, drive the data enable of your
  -- first stage of processing with the signal GD11_CH9_EN.
  GD11_CH9_EN <= (INFIFO_DVALID(0) AND (    INFIFO0_D(15)) AND (NOT INFIFO0_D(14)) AND (NOT INFIFO0_D(13)) AND (    INFIFO0_D(12))) OR
                 (DVALID_DELAYED   AND (     TOP_HALF(15)) AND (NOT  TOP_HALF(14)) AND (NOT  TOP_HALF(13)) AND (     TOP_HALF(12)));

  -- To take data from HEGD11 Channel 10, drive the data enable of your
  -- first stage of processing with the signal GD11_CH10_EN.
  GD11_CH10_EN <= (INFIFO_DVALID(0) AND (    INFIFO0_D(15)) AND (NOT INFIFO0_D(14)) AND (    INFIFO0_D(13)) AND (NOT INFIFO0_D(12))) OR
                  (DVALID_DELAYED   AND (     TOP_HALF(15)) AND (NOT  TOP_HALF(14)) AND (     TOP_HALF(13)) AND (NOT  TOP_HALF(12)));

  -- To take data from HEGD11 Channel 11, drive the data enable of your
  -- first stage of processing with the signal GD11_CH11_EN.
  GD11_CH11_EN <= (INFIFO_DVALID(0) AND (    INFIFO0_D(15)) AND (NOT INFIFO0_D(14)) AND (    INFIFO0_D(13)) AND (    INFIFO0_D(12))) OR
                  (DVALID_DELAYED   AND (     TOP_HALF(15)) AND (NOT  TOP_HALF(14)) AND (     TOP_HALF(13)) AND (     TOP_HALF(12)));

  -- To use unsigned data, connect to the following 12-bit bus
  GD11_UNS_DATA <= INFIFO0_D(11 downto 0) when INFIFO_DVALID(0)='1' else TOP_HALF(11 downto 0);

  -- To use signed data, connect to the following 12-bit bus
  GD11_SGN_DATA <= (NOT GD11_UNS_DATA(11)) & GD11_UNS_DATA(10 downto 0);
